//Created by Kaleb Dempsey
//Last updated: 1/26/24

//This was designed  to record measurements from a Stevens Hydraprobe II and a string pot.

#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Adafruit_GPS.h> //for gps
#include <SDISerial.h>
#include <SD.h>

//User Inputs
unsigned long gps_update_period = 8000; //force GPS update
float dist_m = 2.5025; //Distance calibration factor
float dist_b = -2.3132; //Distance calibration offset

//PIN Definitions
const int Depth_Pin = A8; //string pot analod read
const int chipSelect = 53;// SD card read/write

//String Pot Definitions
float depth = 0; //in

//GPS Related Definitions
#define GPSSerial Serial3
Adafruit_GPS GPS(&GPSSerial);
bool gps_found = false;
bool gps_search = true;
bool gps_on = true;
bool gps_exist = false; //for screen display
unsigned long time_start_gps = millis();// set timer to update GPS at request

//MC probe related
#define DATALINE_PIN 2 //pwm 2
#define INVERTED 3   //pwm 3
SDISerial connection(DATALINE_PIN,INVERTED);

//---------Recorded Data--------------------
bool trigger = false; //write code to set trigger of data collection
char sendBuf[200];
String SDcommand;
char csv_filename[20] = "FILENAME.csv"; //Input desired filename for recording the measurements

//Stevens HydraProbe II Definitions
char *result;//sign replacement function
float soil_temp = 0; //deg C
float mc_content = 0; //VMC
float soil_cond = 0; //mS/cm

//GPS Data Definitions
int year = 10;
int month = 10;
int day = 1;
int hour = 1;
int minute = 1;
int second = 1;;
float latitude = 1.0; //degrees
float longitude = 5.0; //degrees
char time_compiled[200];

//Foward Declerations of Functions
void Trigger_Collection();
void getPenData();
void sendPenData();

void setup() 
  {
    Serial.begin(9600); //Serial Monitor
    connection.begin(); //SDI-12
    GPS.begin(9600);
    GPS.sendCommand(PMTK_SET_NMEA_OUTPUT_RMCONLY);
    GPS.sendCommand(PMTK_SET_NMEA_UPDATE_1HZ);

    if (!SD.begin(chipSelect)) 
      {
        Serial.println("sd card initialization failed!");
        return;
      }
    
    delay(1000);
    char *measurement_data= connection.sdi_query("0XR!",2000); //set mode on SDI-12 sensor
    delay(2000);
    
    //analogReference(EXTERNAL); //Set to Internal if external reference is not used
      
    Serial.write("SystemReady\n"); //Write over serial to know it's ready
  }

char* Replace_Sign(const char* rawdata, const char* old_charc, 
                  const char* new_charc) 
  { 
    //SDI12 is sign delimited ex: 0+1.1-2.3+42.1+5.0
    //we want to be able to parse results so we can print to screen
    //if we don't do this and a "-" exists , an indexed value will be 0,1.1-2.3,42.1,5.0,, (last indexed value empty)
    //instead of 0,1.1,-2.3, 42.1,5.0
    //this function replaces '-' with "+-", which we later parse on "+" to keep our negative signs if they exist

    free(result);
    int i, j = 0; 
    int new_str_len = strlen(new_charc); 
    int old_str_len = strlen(old_charc); 
  
    // Counting the number of times old character occur in the string 
    for (i = 0; rawdata[i] != '\0'; i++) 
    { 
      if (strstr(&rawdata[i], old_charc) == &rawdata[i]) 
          { 
          j++; 
          // Jumping to index after the old character. 
          i += old_str_len - 1; 
          } 
    } 
  
    // Make new string of enough length 
    result = (char*)malloc(100); 
    i = 0; 
    
    while (*rawdata) 
      { 
          // compare the substring with the result 
        if (strstr(rawdata, old_charc) == rawdata) 
          { 
          strcpy(&result[i], new_charc); 
          i += new_str_len; 
          rawdata += old_str_len; 
          } 
          else
              result[i++] = *rawdata++; 
      } 

    result[i] = '\0'; 
    return result; 
  }
  
void loop() 
  {
   unsigned long time_now = millis();
   Trigger_Collection();
   //SD Card communication, retrieve data from SD card   
   if(Serial.available()>2)
    {
      SDcommand = Serial.readStringUntil('\n');
      
      if (SDcommand.equals("SendData"))//Command over serial monitor to pull data off SD if wanted
        {
          File dataFile = SD.open(csv_filename);
  
          // if the file is available, write to it:
          if (dataFile) 
            {
              while (dataFile.available()) 
                {
                  Serial.write(dataFile.read());
                }
              dataFile.close();
            } 
        } 
    }
  
   //routinely check for GPS instead of at measurement
   if (time_now - time_start_gps >= gps_update_period)
      {
       time_start_gps = millis();
       
       if(gps_on && gps_search == false)
        {
          year = int(GPS.year);
          month = int(GPS.month);
          day = int(GPS.day);
          hour = int(GPS.hour);
          minute = int(GPS.minute);
          second = int(GPS.seconds);
          latitude = float(GPS.latitudeDegrees);
          longitude = float(GPS.longitudeDegrees);
        }
      }
    
   if (gps_on == true)
      {
       if(GPS.fix && gps_search == true)
         {
          gps_found = true;
          gps_search = false;
         }
      //Read GPS when available
      while (GPSSerial.available()>0)
         {
            char gps = GPS.read();
            if (GPS.newNMEAreceived()) 
              { 
                if (!GPS.parse(GPS.lastNMEA())) // this also sets the newNMEAreceived() flag to false
                return; // we can fail to parse a sentence in which case we should just wait for another
              }     
          }
      }
  }

void Trigger_Collection(void)
    {   
      //write some condition to trigger data collection
      
      if (trigger == true)
        {
          getPenData();
          sendPenData();
        }
    }

void getPenData(void)
  {      
    //Send SDI-12 measurment request
    //0M! returns sensor address, time to make request, and number parameters
    //ex: returns 0024 is address 0, 2 seconds, returns 4 parameters
    char *measurement_data= connection.sdi_query("0M!",1000);
    delay(2200);

    //recieve the measurement from the probe (at min 2 sec wait)
    char *measurement_data1 = connection.sdi_query("0D0!",1000); // get measurements from probe
    char *sign_change = NULL;
    sign_change = Replace_Sign(measurement_data1,"-","+-");

    //parse the SDI-12 data from the probe
    int index = 0; //index start for parsing
    char *results[100];//read only array for storing parsed results
    char *ptr; //read only array copy of measurement data
    ptr = strtok(sign_change, "+"); //parsing every '+'

    while (ptr != NULL)
      {
        results[index++] = ptr;
        ptr = strtok(NULL, "+");
      }

    //zero index is sensor address and ignored
    soil_temp = atof(results[3]); //see Table A4 F in user manual
    mc_content = atof(results[1])*100; //see Table A4 H in user manual
    soil_cond = atof(results[2]); //see Table A4 J in user manual

    delay(500);
    int raw = analogRead(A8); //read stringpot data
    depth = float(raw)*5/1023;
    depth = depth*dist_m+dist_b;//inch

    if(gps_on == true)
      {
        year = int(GPS.year);
        month = int(GPS.month);
        day = int(GPS.day);
        minute = int(GPS.minute);
        second = int(GPS.seconds);
        latitude = float(GPS.latitudeDegrees);
        longitude = float(GPS.longitudeDegrees);

        if (year==0)
          {
             //To keep formattig consistent for database logging 
             year = 11;
             month = 1;
             day = 1;
             hour = 1;
             minute = 1;
             second = 1;
          }
      }
    
    if (gps_on == false)
      {
        year = 10;
        month = 1;
        day = 1;
        hour = 1;
        minute = 1;
        second = 1;
        latitude = 0.0;
        longitude = 0.0;
      } 

      //data+time string is to large for mesh Xbee packet. 
      //This is setup for storing locally to file in similar database format
      //Note year 2011 indicates GPS is on but no there is no satelite fix
      //year 2010 indicates GPS was turned off locally
      //year 2100 change 20 to 21 :)
      snprintf(time_compiled,sizeof(time_compiled),
                "20%d/%d/%d %d:%d:%d,",
                year,
                month,
                day,
                hour,
                minute,
                second);
  }

void sendPenData(void)
  {
      //compile data to strings and make packets
      char lat_outstr[15];
      char long_outstr[15];
      char depth_outstr[15];
      char temp_outstr[15];
      char mc_outstr[15];
      char cond_outstr[15];

      dtostrf(latitude,4,7,lat_outstr); //float to strings for output
      dtostrf(longitude,4,7,long_outstr); 
      dtostrf(soil_temp,2,1,temp_outstr);
      dtostrf(mc_content,3,1,mc_outstr);
      dtostrf(soil_cond,1,4,cond_outstr);
      dtostrf(depth,2,2,depth_outstr);
      
      snprintf(sendBuf, sizeof(sendBuf),
              "%s,%s,%s,%s,%s,%s\r",
                lat_outstr,
                long_outstr,
                depth_outstr,
                temp_outstr,
                mc_outstr,
                cond_outstr);
                
    //Data always stored to SD Card
    strcat(time_compiled,sendBuf); //merge timestring + data for local storage
    
    //write data to SD
    File dataFile = SD.open(csv_filename, FILE_WRITE);

    if (dataFile) 
      {
        dataFile.print(time_compiled); //write packet to SD
        dataFile.close();
        Serial.println(time_compiled); //print data over serial if connected
      }
    
    //optionally write code to send time_compiled (packet with data collected) elsewhere
  }
